pub fn min_plus_convolution_piecewise_linear<T>(a: &[T], b: &[T]) -> Vec<T>Expand description
Computes convolution using the input with fewer maximal linear pieces.
If that input has p pieces, the running time is O(p * (n + m)).
Panics
Panics unless at least one input is finite, or an index cannot be
represented by T.