MIntConvolve

Type Alias MIntConvolve 

pub type MIntConvolve<M> = Convolve<(M, (Modulo167772161, Modulo469762049, Modulo754974721))>;
Raw transforms require each integer coefficient reconstructed by CRT to be below the product of the three NTT moduli. convolve splits products exceeding this bound.

Aliased Type

pub struct MIntConvolve<M>(PhantomData<fn() -> (M, (Modulo167772161, Modulo469762049, Modulo754974721))>);

Tuple Fields

0: PhantomData<fn() -> (M, (Modulo167772161, Modulo469762049, Modulo754974721))>